__title__ = "mathy_core"
__version__ = "0.8.2"
__summary__ = "Computer Algebra System for working with math expressions"
__uri__ = "https://mathy.ai"
__author__ = "Justin DuJardin"
__email__ = "justin@dujardinconsulting.com"
__license__ = "All rights reserved"
